import {
  AbortException,
  MissingPDFException,
  PasswordException,
  UnexpectedResponseException,
  UnknownErrorException,
} from './util'

enum StreamKind {
  UNKNOWN,
  CANCEL,
  CANCEL_COMPLETE,
  CLOSE,
  ENQUEUE,
  ERROR,
  PULL,
  PULL_COMPLETE,
  START_COMPLETE,
}

enum CallbackKind {
  DATA,
  ERROR,
}

interface PromiseWithOptionalResolvers<T> extends PromiseWithResolvers<T> {
  resolve: (value?: T | PromiseLike<T>) => void;
}

interface errorMsg extends Error {
  message: string
  name: string
  code?: number
  status?: number
  details?: string
}

interface streamMessage {
  targetName: string
  stream: StreamKind
  streamId: number
  success: boolean
  reason: errorMsg
  desiredSize: number
  chunk: any
}
type MessageData = {
  sourceName: string;
  targetName: string;
  action?: string;
  callbackId?: number;
  data?: any;
  callback?: CallbackKind;
  reason?: any;
  stream?: StreamKind;
  streamId?: number;
  desiredSize?: number;
  chunk?: any;
}

type StreamSink = {
  enqueue(chunk: any, size?: number, transfers?: StructuredSerializeOptions): void;
  close(): void;
  error(reason: any): void;
  sinkCapability: PromiseCapability<void>;
  onPull: (() => void) | null;
  onCancel: ((reason: any) => void) | null;
  isCancelled: boolean;
  desiredSize: number;
  ready: Promise<void> | null;
}

type StreamController = {
  controller: ReadableStreamDefaultController<any>;
  startCall: PromiseWithOptionalResolvers<unknown>;
  pullCall: PromiseWithOptionalResolvers<unknown> | null;
  cancelCall: PromiseWithOptionalResolvers<unknown> | null;
  isClosed: boolean;
}

type PromiseCapability<T> = {
  promise: Promise<T>;
  resolve: (value?: T | PromiseLike<T>) => void;
  reject: (reason?: any) => void;
}

function wrapReason(reason: errorMsg) {
  if (!(
    reason instanceof Error ||
    (typeof reason === 'object' && reason !== null && 'message' in reason))
  ) {
    throw new Error('Expected an Error or an object with a message property')
  }

  switch (reason.name) {
    case 'AbortException':
      return new AbortException(reason.message)
    case 'MissingPDFException':
      return new MissingPDFException(reason.message)
    case 'PasswordException':
      return new PasswordException(reason.message, reason.code!)
    case 'UnexpectedResponseException':
      return new UnexpectedResponseException(reason.message, reason.status!)
    case 'UnknownErrorException':
      return new UnknownErrorException(reason.message, reason.details!)
    default:
      return new UnknownErrorException(reason.message, reason.toString())
  }
}

class MessageHandler {
  sourceName: string
  targetName: string
  comObj: MessagePort
  callbackId: number
  streamId: number
  streamSinks: { [key: number]: StreamSink }
  streamControllers: { [key: number]: StreamController }
  callbackCapabilities: { [key: number]: PromiseCapability<any> }
  actionHandler: { [key: string]: Function }
  _onComObjOnMessage: (event: MessageEvent) => void

  constructor(sourceName: string, targetName: string, comObj: any) {
    this.sourceName = sourceName
    this.targetName = targetName
    this.comObj = comObj
    this.callbackId = 1
    this.streamId = 1
    this.streamSinks = Object.create(null)
    this.streamControllers = Object.create(null)
    this.callbackCapabilities = Object.create(null)
    this.actionHandler = Object.create(null)

    this._onComObjOnMessage = event => {
      const data = event.data

      if (data.targetName !== this.sourceName) return

      if (data.stream) {
        this.#processStreamMessage(data)
        return
      }

      if (data.callback) {
        const callbackId = data.callbackId
        const capability = this.callbackCapabilities[callbackId]

        if (!capability) {
          throw new Error(`Cannot resolve callback ${callbackId}`)
        }
        delete this.callbackCapabilities[callbackId]

        if (data.callback === CallbackKind.DATA) {
          capability.resolve(data.data)
        } else if (data.callback === CallbackKind.ERROR) {
          capability.reject(wrapReason(data.reason))
        } else {
          throw new Error('Unexpected callback case')
        }
        return
      }

      const action = this.actionHandler[data.action]
      if (!action) {
        throw new Error(`Unknown action from worker: ${data.action}`)
      }

      if (data.callbackId) {
        const cbSourceName = this.sourceName
        const cbTargetName = data.sourceName

        new Promise((resolve, reject) => {
          try {
            resolve(action(data.data))
          } catch (ex) {
            reject(ex)
          }
        }).then(
          result => {
            comObj.postMessage({
              sourceName: cbSourceName,
              targetName: cbTargetName,
              callback: CallbackKind.DATA,
              callbackId: data.callbackId,
              data: result,
            })
          },
          reason => {
            comObj.postMessage({
              sourceName: cbSourceName,
              targetName: cbTargetName,
              callback: CallbackKind.ERROR,
              callbackId: data.callbackId,
              reason: wrapReason(reason),
            })
          }
        )
        return
      }

      if (data.streamId) {
        this.#createStreamSink(data)
        return
      }
      action(data.data)
    }

    comObj.addEventListener('message', this._onComObjOnMessage)
  }

  on(actionName: string, handler: Function) {
    const actionHandler = this.actionHandler
    if (actionHandler[actionName]) {
      throw new Error(`There is already an actionName called "${actionName}"`)
    }

    actionHandler[actionName] = handler
  }

  send(actionName: string, data: any, transfers?: StructuredSerializeOptions) {
    this.comObj.postMessage({
      sourceName: this.sourceName,
      targetName: this.targetName,
      action: actionName,
      data,
    }, transfers)
  }

  async sendWithPromise(actionName: string, data: any, transfers?: StructuredSerializeOptions) {
    const callbackId = this.callbackId++
    const capability = Promise.withResolvers()
    this.callbackCapabilities[callbackId] = capability

    try {
      this.comObj.postMessage({
        sourceName: this.sourceName,
        tartgetName: this.targetName,
        action: actionName,
        callbackId,
        data,
      },
      transfers)
    } catch (ex) {
      capability.reject(ex)
    }

    return capability.promise
  }

  sendWithStream(actionName: string, data: any, queuingStrategy: QueuingStrategy, transfers?: StructuredSerializeOptions) {
    const streamId = this.streamId++,
      sourceName = this.sourceName,
      targetName = this.targetName,
      comObj = this.comObj

    return new ReadableStream(
      {
        start: controller => {
          const startCabability = Promise.withResolvers()

          this.streamControllers[streamId] = {
            controller,
            startCall: startCabability,
            pullCall: null,
            cancelCall: null,
            isClosed: false,
          }

          comObj.postMessage({
            sourceName,
            targetName,
            action: actionName,
            streamId,
            data,
            deserialize: controller.desiredSize,
          }, transfers)

          return startCabability.promise
        },

        pull: controller => {
          const pullCabability = Promise.withResolvers()

          this.streamControllers[streamId].pullCall = pullCabability

          comObj.postMessage({
            sourceName,
            targetName,
            stream: StreamKind.PULL,
            streamId,
            desiredSize: controller.desiredSize,
          })

          return pullCabability.promise as PromiseLike<void>
        },

        cancel: reason => {
          if (!(reason instanceof Error)) {
            throw new Error('cancel must have a valid reason');
          }

          const cancelCabability = Promise.withResolvers()
          this.streamControllers[streamId].cancelCall = cancelCabability
          this.streamControllers[streamId].isClosed = true

          comObj.postMessage({
            sourceName,
            targetName,
            stream: StreamKind.CANCEL,
            streamId,
            reason: wrapReason(reason),
          })

          return cancelCabability.promise as PromiseLike<void>
        }
      },
      queuingStrategy
    )
  }

  #createStreamSink(data: MessageData) {
    const streamId = data.streamId,
      sourceName = this.sourceName,
      targetName = this.targetName,
      comObj = this.comObj,
      self = this,
      action = this.actionHandler[data.action!]

    const streamSink: StreamSink = {
      enqueue(chunk, size = 1, transfers?: StructuredSerializeOptions) {
        if (this.isCancelled) {
          return
        }

        const lastDesiredSize = this.desiredSize
        this.desiredSize -= size

        if (lastDesiredSize > 0 && this.desiredSize <= 0) {
          this.sinkCapability = Promise.withResolvers()
          this.ready = this.sinkCapability.promise
        }

        comObj.postMessage({
          sourceName,
          targetName,
          stream: StreamKind.ENQUEUE,
          streamId,
          chunk,
        }, transfers)
      },

      close() {
        if (this.isCancelled) {
          return
        }

        this.isCancelled = true

        comObj.postMessage({
          sourceName,
          targetName,
          stream: StreamKind.CLOSE,
          streamId,
        })
        streamId && delete self.streamSinks[streamId]
      },

      error(reason) {
        if (!(reason instanceof Error)) {
          throw new Error('error must have a valid reason')
        }

        this.isCancelled = true

        comObj.postMessage({
          sourceName,
          targetName,
          stream: StreamKind.ERROR,
          streamId,
          reason: wrapReason(reason),
        })
      },

      sinkCapability: Promise.withResolvers(),
      onPull: null,
      onCancel: null,
      isCancelled: false,
      desiredSize: data.desiredSize??0,
      ready: null,
    }

    streamSink.sinkCapability.resolve()
    streamSink.ready = streamSink.sinkCapability.promise
    streamId && (this.streamSinks[streamId] = streamSink)

    new Promise(resolve => {
      resolve(action(data.data, streamSink))
    }).then(
      () => {
        comObj.postMessage({
          sourceName,
          targetName,
          stream: StreamKind.START_COMPLETE,
          streamId,
          success: true,
        })
      },
      reason => {
        comObj.postMessage({
          sourceName,
          targetName,
          stream: StreamKind.START_COMPLETE,
          streamId,
          reason: wrapReason(reason),
        })
      }
    )
  }

  #processStreamMessage(data: streamMessage) {
    const streamId = data.streamId,
      sourceName = this.sourceName,
      targetName = this.sourceName,
      comObj = this.comObj,
      streamController = this.streamControllers![streamId],
      streamSink = this.streamSinks![streamId]

    switch (data.stream) {
      case StreamKind.START_COMPLETE:
        if (data.success) {
          streamController.startCall.resolve()
        } else {
          streamController.startCall.reject(wrapReason(data.reason))
        }
        break
      case StreamKind.PULL_COMPLETE:
        if (data.success) {
          streamController.pullCall!.resolve()
        } else {
          streamController.pullCall!.reject(wrapReason(data.reason))
        }
        break
      case StreamKind.PULL:
        if (!streamSink) {
          comObj.postMessage({
            sourceName,
            targetName,
            stream: StreamKind.PULL_COMPLETE,
            streamId,
            success: true,
          })
          break
        }

        if (streamSink.desiredSize <= 0 && data.desiredSize > 0) {
          streamSink.sinkCapability.resolve()
        }

        streamSink.desiredSize = data.desiredSize

        new Promise(resolve => {
          resolve(streamSink.onPull?.())
        }).then(
          () => {
            comObj.postMessage({
              sourceName,
              targetName,
              stream: StreamKind.PULL_COMPLETE,
              streamId,
              success: true,
            })
          },
          reason => {
            comObj.postMessage({
              sourceName,
              targetName,
              stream: StreamKind.PULL_COMPLETE,
              streamId,
              reason: wrapReason(reason),
            })
          }
        )
        break
      case StreamKind.ENQUEUE:
        if (!streamController) {
          throw new Error('enqueue should have stream controller')
        }
        if (streamController.isClosed) {
          break
        }

        streamController.controller.enqueue(data.chunk)
        break
      case StreamKind.CLOSE:
        if (!streamController) {
          throw new Error('close should have stream controller')
        }
        if (streamController.isClosed) {
          break
        }

        streamController.isClosed = true
        streamController.controller.close()
        this.#deleteStreamController(streamController, streamId)
        break
      case StreamKind.ERROR:
        if (!streamController) {
          throw new Error('error should have stream controller')
        }

        streamController.controller.error(wrapReason(data.reason))
        this.#deleteStreamController(streamController, streamId)
        break
      case StreamKind.CANCEL:
        if (!streamSink) {
          break
        }

        new Promise(resolve => {
          resolve(streamSink.onCancel?.(wrapReason(data.reason)))
        }).then(
          () => {
            comObj.postMessage({
              sourceName,
              targetName,
              stream: StreamKind.CANCEL_COMPLETE,
              streamId,
              success: true,
            })
          },
          reason => {
            comObj.postMessage({
              sourceName,
              targetName,
              stream: StreamKind.CANCEL_COMPLETE,
              streamId,
              reason: wrapReason(reason),
            })
          }
        )
        streamSink.sinkCapability.reject(wrapReason(data.reason))
        streamSink.isCancelled = true
        delete this.streamSinks[streamId]
        break
      default:
        throw new Error(`Unexpected stream case`)
    }
  }

  async #deleteStreamController(streamController: StreamController, streamId: number) {
    await Promise.allSettled([
      streamController.startCall?.promise,
      streamController.pullCall?.promise,
      streamController.cancelCall?.promise,
    ]);
    delete this.streamControllers[streamId]
  }

  destroy() {
    this.comObj.removeEventListener("message", this._onComObjOnMessage);
  }
}

export default MessageHandler
